package filegateway

import (
	"context"
	"fmt"
	"io"
	"path"

	nrpc "github.com/nats-rpc/nrpc"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	"orus.io/xbus/xbus-stdlib/service"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"
)

/*
```yaml
actor:
  kind: consumer
  service:
    type: file-gateway
    settings:
       uri: scp://remote/directory/
       secrets: path/to/passwords-file.yaml
       poll-delay: 10s
       wait-for: deletion
```
*/

// Consumer save the received files to a filesystem.
type Consumer struct {
	fs        Filesystem
	targetDir string

	temporaryErrors bool
}

// NewConsumer creates a Consumer.
func NewConsumer(actor *api.Actor, settings service.Settings, log xbus.Logger) (service.Consumer, error) {
	options, err := LoadFilesystemOptions(settings, "target")
	if err != nil {
		return nil, fmt.Errorf("could not load 'target' settings: %w", err)
	}
	fs, err := GetFilesystem(options.Scheme, options, log.AsZero())
	if err != nil {
		return nil, fmt.Errorf("could not get filesystem: %w", err)
	}

	temporaryErrors, err := settings.BoolD("temporary-errors", false)
	if err != nil {
		return nil, fmt.Errorf("could not read 'temporary-errors': %w", err)
	}
	return &Consumer{
		fs:              fs,
		targetDir:       options.Path,
		temporaryErrors: temporaryErrors,
	}, nil
}

func (c *Consumer) fsError(op string, targetPath string, err error) error {
	if c.temporaryErrors {
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[filegateway] Error %s file '%s': %s.", op, targetPath, err),
		}
	}

	return err
}

// Process ...
func (c *Consumer) Process(
	ctx context.Context, request *api.ActorProcessRequest, inputs service.Inputs, log xbus.Logger,
) error {
	defaultInput, err := inputs.Get("default")
	if err != nil {
		return err
	}

	if err := envelope.ForEachMessage(
		defaultInput,
		func(msg envelope.Message) error {
			fileReader := stdlib.NewFileMessageReader(msg)

			meta, err := fileReader.Metadata()
			if err != nil {
				return err
			}
			body, err := fileReader.Body()
			if err != nil {
				return err
			}
			defer func() {
				if err := body.Close(); err != nil {
					log.Err(err).Msg("error closing the file body")
				}
			}()
			targetPath := path.Join(c.targetDir, meta.Filename)
			log.Debug().Str("target-path", targetPath).Str("fs", c.fs.String()).Msg("Writing")
			writer, err := c.fs.Write(targetPath, -1)
			if err != nil {
				return c.fsError("opening", targetPath, err)
			}
			if _, err = io.Copy(writer, body); err != nil {
				if err := writer.Close(); err != nil {
					log.Warn().Err(err).Msg("error closing the writer during an error handling")
				}
				return c.fsError("writing", targetPath, err)
			}
			if err := writer.Close(); err != nil {
				return c.fsError("closing", targetPath, err)
			}
			log.Info().
				Str("target-path", targetPath).
				Str("fs", c.fs.String()).
				Msgf("Successfully written %s to %s", targetPath, c.fs)

			return nil
		},
	); err != nil {
		log.Err(err).Msg("filegateway consumer error")
		return err
	}

	return nil
}

func init() {
	service.MustRegisterConsumer("file-gateway-consumer", NewConsumer)
}
