package filegateway

import (
	"errors"
	"fmt"
	"os"
	"sync/atomic"
	"time"

	"github.com/pkg/sftp"
	"golang.org/x/crypto/ssh"
)

var (
	ErrSFtpPoolFull = errors.New("sftp pool is full")
)

// NewSFtpPool creates a sftp connection pool
func NewSFtpPool(address string, config *ssh.ClientConfig, maxsize int32, keepAlive time.Duration) *SFtpPool {
	return &SFtpPool{
		address:   address,
		config:    *config,
		maxsize:   maxsize,
		keepAlive: keepAlive,
		count:     0,
		conns:     make(chan *sftpPoolConn, maxsize),
	}
}

// SFtpPool is a pool of sftp connections
type SFtpPool struct {
	address   string
	config    ssh.ClientConfig
	maxsize   int32
	keepAlive time.Duration

	count int32
	conns chan *sftpPoolConn
}

type sftpPoolConn struct {
	client *sftp.Client
	timer  *time.Timer
}

func (p *SFtpPool) createConn() (*sftp.Client, error) {
	conn, err := ssh.Dial("tcp", p.address, &p.config)
	if err != nil {
		return nil, fmt.Errorf("Failed to dial: %s", err)
	}
	client, err := sftp.NewClient(conn)
	return client, err
}

func (p *SFtpPool) add() (*sftp.Client, error) {
	count := atomic.LoadInt32(&p.count)
	if count == p.maxsize {
		return nil, ErrSFtpPoolFull
	}
	if atomic.CompareAndSwapInt32(&p.count, count, count+1) {
		client, err := p.createConn()
		if err != nil {
			atomic.AddInt32(&p.count, -1)
		}
		return client, err
	}
	return nil, nil
}

func (p *SFtpPool) isFull() bool {
	return atomic.LoadInt32(&p.count) == p.maxsize
}

func (p *SFtpPool) prepareConn(c *sftpPoolConn) *sftp.Client {
	if c.timer.Stop() {
		return c.client
	}
	atomic.AddInt32(&p.count, -1)
	return nil
}

// Get returns a sftp connection
func (p *SFtpPool) Get() (*sftp.Client, error) {
	for {
		if p.isFull() { // nolint: nestif
			c := <-p.conns
			client := p.prepareConn(c)
			if client != nil {
				return client, nil
			}
		} else {
			select {
			case c := <-p.conns:
				client := p.prepareConn(c)
				if client != nil {
					return client, nil
				}
			default:
				client, err := p.add()
				if err != nil {
					if errors.Is(err, ErrSFtpPoolFull) {
						continue
					}
					return nil, err
				}
				if client != nil {
					return client, nil
				}
			}
		}
	}
}

// Release put back a connection to the pool
func (p *SFtpPool) Release(client *sftp.Client) {
	conn := sftpPoolConn{
		client: client,
		timer:  nil,
	}
	conn.timer = time.AfterFunc(p.keepAlive, func() {
		conn.client.Close()
		conn.client = nil
	})
	p.conns <- &conn
}

// Drop releases a connection and removes from the pool
func (p *SFtpPool) Drop(client *sftp.Client) {
	atomic.AddInt32(&p.count, -1)
	client.Close()
}

// Close closes all the connections
func (p *SFtpPool) Close() error {
	return nil
}

// Stat runs Stat on a connection
func (p *SFtpPool) Stat(path string) (info os.FileInfo, err error) {
	err = p.Exec(func(client *sftp.Client) error {
		info, err = client.Stat(path)
		return err
	})
	return
}

// Glob runs Glob on a connection
func (p *SFtpPool) Glob(pattern string) (matches []string, err error) {
	err = p.Exec(func(client *sftp.Client) error {
		matches, err = client.Glob(pattern)
		return err
	})
	return
}

// Exec runs a function on a sftp client.
func (p *SFtpPool) Exec(run func(client *sftp.Client) error) error {
	for {
		conn, err := p.Get()
		if err != nil {
			return err
		}
		err = run(conn)
		if err == sftp.ErrSshFxConnectionLost {
			p.Drop(conn)
		} else {
			p.Release(conn)
			return err
		}
	}
}
